inline bool rayIntersectsTriangle(float3 p, float3 sunDirection, float3 v0, float3 v1, float3 v2) {
	float3 e1 = v1 - v0;
	float3 e2 = v2 - v0;
			
	float3 h = cross(sunDirection, e2);
			
	float a = dot(e1, h);
			
	if ((a > -0.00001f) && (a < 0.00001f)) {
		return false;
	}
			
	float f = 1 / a;
	
	float3 s2 = p - v0;
	
	float u = f * dot(s2, h);
	
	if ((u < 0.0f) || (u > 1.0f)) {
		return false;
	}
	
	float3 q = cross(s2, e1);
	
	float v = f * dot(sunDirection, q);
	if (!((v < 0.0f) || ((u + v) > 1.0f))) {
		float3 uv = v0 + e1 * u + e2 * v;
		float3 s = uv - p;
			
		if (dot(s, sunDirection) > 0.01f) {
			return true;
		}
	}


	return false;
}

__kernel  
void calc(
	__constant const float* neighborsVertices,
	__constant const float* shadowVerticeCenters,
	__constant const float* shadowTriangleNormals,
	__constant const float* sunDirections,
	__global char* hasShadow, // skymodel * workSize
	__const int workSize,
	__const int skymodel,
	__const int neighborWorksize)
{
	int gid = get_global_id(0);
	if (gid >= workSize) {
		return;
	}
	
	float3 v0, v1, v2, sunDirection, p, n;
	float l1;

	int gid3 = gid*3; 
	p = (float3)(shadowVerticeCenters[gid3], shadowVerticeCenters[gid3+1], shadowVerticeCenters[gid3+2]);
	n = (float3)(shadowTriangleNormals[gid3], shadowTriangleNormals[gid3+1], shadowTriangleNormals[gid3+2]);
	l1 = fast_length(n);
		 
	for (int i = 0; i < skymodel; ++i) {
		int hasShadowIndex = gid * skymodel + i;
		hasShadow[hasShadowIndex] = 0;
		if (hasShadow[hasShadowIndex] != 0) {
			continue;
		}
		sunDirection = (float3)(sunDirections[i*3], sunDirections[i*3+1], sunDirections[i*3+2]);
		float l0 = fast_length(sunDirection);
		float c = dot(sunDirection, n) / (l0 * l1);
		if (c < 0) {
			hasShadow[hasShadowIndex] = 1;
			continue;
		}
		for (int neighborIdx = 0; neighborIdx < neighborWorksize; ++neighborIdx) {
			int j = neighborIdx * 9;
			v0 = (float3)(neighborsVertices[j+0], neighborsVertices[j+1], neighborsVertices[j+2]);
			v1 = (float3)(neighborsVertices[j+3], neighborsVertices[j+4], neighborsVertices[j+5]);
			v2 = (float3)(neighborsVertices[j+6], neighborsVertices[j+7], neighborsVertices[j+8]);
			bool res = rayIntersectsTriangle(p, sunDirection, v0, v1, v2);
			if (res) {
				hasShadow[hasShadowIndex] = 1;
				break;
			}
		}
	}  
}